/*
 * LCD peripheral support for the JZ4740 and related SoCs.
 *
 * Copyright (C) Xiangfu Liu <xiangfu@sharism.cc>
 * Copyright (C) 2015, 2016, 2017, 2018,
 *               2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/hw_mmio_register_block.h>
#include <l4/sys/cache.h>
#include <l4/sys/irq.h>
#include <l4/sys/types.h>
#include <l4/util/util.h>

#include "lcd-jz4740.h"
#include "lcd-jz4740-config.h"
#include "lcd-jz4740-regs.h"

#include <stdint.h>



// Utility functions.

// Round values up according to the resolution.

static uint32_t align(uint32_t value, uint32_t resolution)
{
  return (value + (resolution - 1)) & ~(resolution - 1);
}

// Value pair encoding.

static uint32_t encode_pair(uint32_t start, uint32_t end)
{
  return (start << Value_first) | (end << Value_second);
}

// RGB conversions.

static uint16_t rgb8_to_rgb16(uint8_t rgb)
{
  return ((((rgb & 0xe0) >> 5) * 4) << 11) | ((((rgb & 0x1c) >> 2) * 9) << 6) | ((rgb & 0x03) * 10);
}

static uint16_t rgb4_to_rgb16(uint8_t rgb)
{
  return ((((rgb & 8) >> 3) * 0x1f) << 11) | ((((rgb & 6) >> 1) * 0x15) << 5) | ((rgb & 1) * 0x1f);
}




// If implemented as a Hw::Device, various properties would be
// initialised in the constructor and obtained from the device tree
// definitions.

Lcd_jz4740_chip::Lcd_jz4740_chip(l4_addr_t addr, Jz4740_lcd_panel *panel)
: _panel(panel)
{
  _regs = new Hw::Mmio_register_block<32>(addr);
  _burst_size = 16; // 16-word burst size
}

struct Jz4740_lcd_panel *
Lcd_jz4740_chip::get_panel()
{
  return _panel;
}

void
Lcd_jz4740_chip::disable()
{
  // Set the disable bit for normal shutdown.

  _regs[Lcd_control] = _regs[Lcd_control] | (1U << Control_disable);
}

void
Lcd_jz4740_chip::disable_quick()
{
  // Clear the enable bit for quick shutdown.

  _regs[Lcd_control] = _regs[Lcd_control] & ~(1U << Control_enable);
}

void
Lcd_jz4740_chip::enable()
{
  // Clear the disable bit and set the enable bit.

  _regs[Lcd_status] = 0;
  _regs[Lcd_control] = (_regs[Lcd_control] & ~(1U << Control_disable)) | (1U << Control_enable);
}

bool
Lcd_jz4740_chip::enabled()
{
  return !(_regs[Lcd_status] & (1U << Status_disabled));
}

// Calculate and return the pixel clock frequency.

int
Lcd_jz4740_chip::get_pixel_clock()
{
  int pclk, multiplier;

  // Serial mode: 3 pixel clock cycles per pixel (one per channel).
  // Parallel mode: 1 pixel clock cycle per pixel.

  multiplier = have_serial_tft() ? 3 : 1;

  // Derive pixel clock rate from frame rate.
  // This multiplies the number of pixel periods in a line by the number of
  // lines in a frame, thus obtaining the number of such periods in a frame.
  // Multiplying this result with the frame rate yields the pixel frequency.

  pclk = _panel->frame_rate *
         (_panel->width * multiplier +
          _panel->hsync + _panel->line_start + _panel->line_end) *
         (_panel->height +
          _panel->vsync + _panel->frame_start + _panel->frame_end);

  // STN panel adjustments.

  if (have_stn_panel())
  {
    // Colour STN panels apparently need to be driven at three times the rate.

    if (have_colour_stn()) pclk = (pclk * 3);

    // Reduce the rate according to the width of the STN connection.
    // Since the pins setting employs log2(pins), a shift by this value is
    // equivalent to a division by the number of pins.

    pclk = pclk >> ((_panel->config & Config_stn_pins_mask) >> Jz4740_lcd_config_stn_pins);

    // Divide the rate by the number of panels.

    pclk /= get_panels();
  }

  return pclk;
}



// Return the panel mode.

uint32_t
Lcd_jz4740_chip::_mode()
{
  return _panel->config & Config_mode_mask;
}

// Return the number of panels available.

int
Lcd_jz4740_chip::get_panels()
{
  uint32_t mode = _mode();

  return (mode == Jz4740_lcd_mode_stn_dual_colour) ||
         (mode == Jz4740_lcd_mode_stn_dual_mono) ? 2 : 1;
}

// Return whether the panel is STN.

int
Lcd_jz4740_chip::have_stn_panel()
{
  uint32_t mode = _mode();

  return ((mode == Jz4740_lcd_mode_stn_single_colour) ||
          (mode == Jz4740_lcd_mode_stn_dual_colour) ||
          (mode == Jz4740_lcd_mode_stn_single_mono) ||
          (mode == Jz4740_lcd_mode_stn_dual_mono));
}

// Return whether the panel is colour STN.

int
Lcd_jz4740_chip::have_colour_stn()
{
  uint32_t mode = _mode();

  return ((mode == Jz4740_lcd_mode_stn_single_colour) ||
          (mode == Jz4740_lcd_mode_stn_dual_colour));
}

// Return whether the panel is colour STN.

int
Lcd_jz4740_chip::have_serial_tft()
{
  return _mode() == Jz4740_lcd_mode_tft_serial;
}



// Return the pixel memory size in bits.

l4_size_t
Lcd_jz4740_chip::get_pixel_size()
{
  if (_panel->bpp > 16)
    return 32;
  else if (_panel->bpp > 8)
    return 16;
  else
    return _panel->bpp;
}

// Return the line memory size.

l4_size_t
Lcd_jz4740_chip::get_line_size()
{
  // Lines must be aligned to a word boundary.

  return align((_panel->width * get_pixel_size()) / 8, sizeof(uint32_t));
}

// Return the screen memory size.

l4_size_t
Lcd_jz4740_chip::get_screen_size()
{
  return get_line_size() * _panel->height;
}

// Return the aligned size for the DMA transfer.

l4_size_t 
Lcd_jz4740_chip::get_aligned_size()
{
  return align(get_screen_size(), _burst_size * sizeof(uint32_t));
}

// Return the size of the palette.

l4_size_t
Lcd_jz4740_chip::get_palette_size()
{
  // No palette for modes with more than eight bits per pixel.

  if (_panel->bpp > 8) return 0;

  // Get the size of a collection of two-byte entries, one per colour.

  return (1U << (_panel->bpp)) * sizeof(uint16_t);
}

// Return the aligned size of the palette for the DMA transfer.

l4_size_t
Lcd_jz4740_chip::get_aligned_palette_size()
{
  return align(get_palette_size(), _burst_size * sizeof(uint32_t));
}

// Return the total memory requirements of the framebuffers and palette.

l4_size_t
Lcd_jz4740_chip::get_total_size()
{
  return get_aligned_size() * get_panels() + get_aligned_palette_size();
}

// Return the total memory requirements of any DMA descriptors.

l4_size_t
Lcd_jz4740_chip::get_descriptors_size()
{
   return 3 * sizeof(struct Jz4740_lcd_descriptor);
}



// Functions returning addresses of each data region.
// The base parameter permits the retrieval of virtual or physical addresses.

l4_addr_t
Lcd_jz4740_chip::get_palette(l4_addr_t base)
{       
  // Use memory at the end of the allocated region for the palette.
        
  return base + (get_panels() * get_aligned_size()) - get_aligned_palette_size();
}

l4_addr_t
Lcd_jz4740_chip::get_framebuffer(int panel, l4_addr_t base)
{
  // Framebuffers for panels are allocated at the start of the region.

  return base + (panel * get_aligned_size());
}



// Palette initialisation.

void
Lcd_jz4740_chip::init_palette(l4_addr_t palette)
{
  uint8_t colours = 1U << (_panel->bpp);
  uint16_t *entry = (uint16_t *) palette;
  uint16_t *end = entry + colours;
  uint8_t value = 0;

  while (entry < end)
  {
    switch (_panel->bpp)
    {
      case 4:
      *entry = rgb4_to_rgb16(value);
      break;

      case 8:
      *entry = rgb8_to_rgb16(value);
      break;

      default:
      break;
    }

    value++;
    entry++;
  }
}



// Return colour depth control value.
// NOTE: Not supporting JZ4780 options.

uint32_t
Lcd_jz4740_chip::_control_bpp()
{
  switch (_panel->bpp)
  {
    case 1:         return Control_bpp_1bpp;
    case 2:         return Control_bpp_2bpp;
    case 3 ... 4:   return Control_bpp_4bpp;
    case 5 ... 8:   return Control_bpp_8bpp;
    case 9 ... 15:  return Control_bpp_15bpp | (Rgb_mode_555 << Control_rgb_mode);
    case 17 ... 18: return Control_bpp_18bpp;
    case 19 ... 32: return Control_bpp_24bpp;
    case 16:
    default:        return Control_bpp_16bpp;
  }
}

// Return a panel-related control value.

uint32_t
Lcd_jz4740_chip::_control_panel()
{
  if (have_stn_panel())
    return _control_stn_frc() << Control_frc_algorithm;
  else
    return 0;
}

// Return a STN-related control value.

uint32_t
Lcd_jz4740_chip::_control_stn_frc()
{
  if (_panel->bpp <= 2)
    return Frc_greyscales_2;
  if (_panel->bpp <= 4)
    return Frc_greyscales_4;
  return Frc_greyscales_16;
}

// Return a transfer-related control value.

uint32_t
Lcd_jz4740_chip::_control_transfer()
{
  uint32_t length;

  switch (_burst_size)
  {
    case 4:  length = Burst_length_4;  break;
    case 8:  length = Burst_length_8;  break;
    case 32: length = Burst_length_32; break;
    case 64: length = Burst_length_64; break;
    case 16:
    default: length = Burst_length_16; break;
  }

  // NOTE: Underrun supposedly not needed.

  return (length << Control_burst_length) | (1U << Control_out_underrun);
}

// Return an interrupt-related control value.

uint32_t
Lcd_jz4740_chip::_control_irq()
{
  return ((_irq_conditions & Lcd_irq_frame_start) ? (1U << Control_frame_start_irq_enable) : 0) |
         ((_irq_conditions & Lcd_irq_frame_end) ? (1U << Control_frame_end_irq_enable) : 0);
}

// Return an interrupt-related command value.

uint32_t
Lcd_jz4740_chip::_command_irq()
{
  return ((_irq_conditions & Lcd_irq_frame_start) ? (1U << Command_frame_start_irq) : 0) |
         ((_irq_conditions & Lcd_irq_frame_end) ? (1U << Command_frame_end_irq) : 0);
}

// Return an interrupt-related status value.

uint32_t
Lcd_jz4740_chip::_status_irq()
{
  return ((_irq_conditions & Lcd_irq_frame_start) ? (1U << Status_frame_start_irq) : 0) |
         ((_irq_conditions & Lcd_irq_frame_end) ? (1U << Status_frame_end_irq) : 0);
}

// STN panel-specific initialisation.

void
Lcd_jz4740_chip::_init_stn()
{
  // Divide the height by the number of panels.

  uint32_t height = _panel->height / get_panels();

  // Since the value is log2(pins), 1 << value yields the number of pins.

  int pins = 1 << ((_panel->config & Config_stn_pins_mask) >> Jz4740_lcd_config_stn_pins);

  // Round parameters up to a multiple of the number of pins.

  uint32_t hsync = align(_panel->hsync, pins);
  uint32_t line_start = align(_panel->line_start, pins);
  uint32_t line_end = align(_panel->line_end, pins);

  // Define the start and end positions of visible data on a line and in a frame.
  // Visible frame data is anchored at line zero, with the start region
  // preceding this line (and thus appearing at the end of the preceding frame).

  uint32_t line_start_pos = line_start;
  uint32_t line_end_pos = line_start_pos + _panel->width;
  uint32_t frame_start_pos = 0;
  uint32_t frame_end_pos = frame_start_pos + height;

  // Define sync pulse locations, with hsync occurring after the visible data.

  _regs[Lcd_hsync] = encode_pair(line_end_pos, line_end_pos + hsync);
  _regs[Lcd_vsync] = encode_pair(0, _panel->vsync);

  // Set the display area and limits.

  _regs[Virtual_area] = encode_pair(line_end_pos + hsync + line_end,
                                    frame_end_pos + _panel->vsync + _panel->frame_end + _panel->frame_start);

  _regs[Display_hlimits] = encode_pair(line_start_pos, line_end_pos);
  _regs[Display_vlimits] = encode_pair(frame_start_pos, frame_end_pos);

  // Set the AC bias signal.

  _regs[Lcd_ps] = encode_pair(0, _panel->frame_start + height + _panel->vsync + _panel->frame_end);
}

// TFT panel-specific initialisation.

void
Lcd_jz4740_chip::_init_tft()
{
  // Define the start and end positions of visible data on a line and in a frame.

  uint32_t line_start_pos = _panel->line_start + _panel->hsync;
  uint32_t line_end_pos = line_start_pos + _panel->width;
  uint32_t frame_start_pos = _panel->frame_start + _panel->vsync;
  uint32_t frame_end_pos = frame_start_pos + _panel->height;

  // Define sync pulse locations, with pulses appearing before visible data.

  _regs[Lcd_hsync] = encode_pair(0, _panel->hsync);
  _regs[Lcd_vsync] = encode_pair(0, _panel->vsync);

  // Set the display area and limits.

  _regs[Virtual_area] = encode_pair(line_end_pos + _panel->line_end,
                                    frame_end_pos + _panel->frame_end);

  _regs[Display_hlimits] = encode_pair(line_start_pos, line_end_pos);
  _regs[Display_vlimits] = encode_pair(frame_start_pos, frame_end_pos);
}

// Initialise the panel.
// NOTE: Only generic STN and TFT panels are supported.

void
Lcd_jz4740_chip::_init_panel()
{
  if (have_stn_panel())
    _init_stn();
  else
    switch (_mode())
    {
      case Jz4740_lcd_mode_tft_generic:
      case Jz4740_lcd_mode_tft_casio:
      case Jz4740_lcd_mode_tft_serial: _init_tft();

      default: break;
    }
}

// Initialise a DMA descriptor.

void
Lcd_jz4740_chip::_set_descriptor(struct Jz4740_lcd_descriptor &desc,
                                 l4_addr_t source, l4_size_t size,
                                 struct Jz4740_lcd_descriptor *next,
                                 uint32_t flags)
{
  // In the command, indicate the number of words from the source for transfer.

  desc.next = next;
  desc.source = source;
  desc.identifier = source;
  desc.command = ((size / sizeof(uint32_t)) & Command_buffer_length_mask) |
                 flags;
}

// Initialise the LCD controller with the memory, panel and framebuffer details.
// Any palette must be initialised separately using get_palette and init_palette.

void
Lcd_jz4740_chip::config(struct Jz4740_lcd_descriptor *desc_vaddr,
                        struct Jz4740_lcd_descriptor *desc_paddr,
                        l4_addr_t fb_paddr)
{
  int have_palette = (_panel->bpp <= 8);

  // Provide the first framebuffer descriptor in single and dual modes.
  // Flip back and forth between any palette and the framebuffer.

  _set_descriptor(desc_vaddr[0], get_framebuffer(0, fb_paddr),
                                 get_aligned_size(),
                                 have_palette ? desc_paddr + 2 : desc_paddr,
                                 _command_irq());

  // Provide the second framebuffer descriptor only in dual-panel mode.
  // Only employ this descriptor in the second DMA channel.

  if (get_panels() == 2)
    _set_descriptor(desc_vaddr[1], get_framebuffer(1, fb_paddr),
                                   get_aligned_size(),
                                   desc_paddr + 1,
                                   _command_irq());

  // Initialise palette descriptor details for lower colour depths.

  if (have_palette)
    _set_descriptor(desc_vaddr[2], get_palette(fb_paddr),
                                   get_aligned_palette_size(),
                                   desc_paddr,
                                   Command_palette_buffer);

  // Flush cached structure data.

  l4_cache_clean_data((unsigned long) desc_vaddr,
                      (unsigned long) desc_vaddr + get_descriptors_size());

  // Configure DMA by setting frame descriptor addresses.

  // Provide the palette descriptor address first, if employed.

  _regs[Desc_address_0] = (uint32_t) (have_palette ? desc_paddr + 2 : desc_paddr);

  // Provide a descriptor for the second DMA channel in dual-panel mode.

  if (get_panels() == 2)
    _regs[Desc_address_1] = (uint32_t) (desc_paddr + 1);

  // Initialise panel-related registers.

  _init_panel();

  // Initialise the control and configuration registers.

  _regs[Lcd_control] = _control_panel() | _control_bpp() | _control_transfer() | _control_irq();
  _regs[Lcd_config] = _panel->config;
}

// Set the interrupt for controller-related events.

void
Lcd_jz4740_chip::set_irq(l4_cap_idx_t irq, enum Jz4740_lcd_irq_condition conditions)
{
  _irq = irq;
  _irq_conditions = conditions;
}

// Wait for an interrupt condition.

long
Lcd_jz4740_chip::wait_for_irq()
{
  l4_msgtag_t tag;

  _regs[Lcd_status] = _regs[Lcd_status] & ~(_status_irq());

  // Wait for a condition.

  tag = l4_irq_receive(_irq, l4_timeout(L4_IPC_TIMEOUT_NEVER, l4util_micros2l4to(2000000)));

  // Acknowledge interrupts.

  _regs[Lcd_status] = 0;

  // Return errors.

  return l4_error(tag);
}



// C language interface functions.

void *
jz4740_lcd_init(l4_addr_t lcd_base, struct Jz4740_lcd_panel *panel)
{
  return (void *) new Lcd_jz4740_chip(lcd_base, panel);
}

void
jz4740_lcd_config(void *lcd, struct Jz4740_lcd_descriptor *desc_vaddr,
                             struct Jz4740_lcd_descriptor *desc_paddr,
                             l4_addr_t fb_paddr)
{
  static_cast<Lcd_jz4740_chip *>(lcd)->config(desc_vaddr, desc_paddr, fb_paddr);
}

void
jz4740_lcd_set_irq(void *lcd, l4_cap_idx_t irq, enum Jz4740_lcd_irq_condition conditions)
{
  static_cast<Lcd_jz4740_chip *>(lcd)->set_irq(irq, conditions);
}

long
jz4740_lcd_wait_for_irq(void *lcd)
{
  return static_cast<Lcd_jz4740_chip *>(lcd)->wait_for_irq();
}

void
jz4740_lcd_disable(void *lcd)
{
  static_cast<Lcd_jz4740_chip *>(lcd)->disable();
}

void
jz4740_lcd_disable_quick(void *lcd)
{
  static_cast<Lcd_jz4740_chip *>(lcd)->disable_quick();
}

void
jz4740_lcd_enable(void *lcd)
{
  static_cast<Lcd_jz4740_chip *>(lcd)->enable();
}

int
jz4740_lcd_enabled(void *lcd)
{
  return (int) static_cast<Lcd_jz4740_chip *>(lcd)->enabled();
}

int
jz4740_lcd_get_pixel_clock(void *lcd)
{
  return static_cast<Lcd_jz4740_chip *>(lcd)->get_pixel_clock();
}

l4_size_t
jz4740_lcd_get_descriptors_size(void *lcd)
{
  return static_cast<Lcd_jz4740_chip *>(lcd)->get_descriptors_size();
}

l4_size_t
jz4740_lcd_get_line_size(void *lcd)
{
  return static_cast<Lcd_jz4740_chip *>(lcd)->get_line_size();
}

l4_size_t
jz4740_lcd_get_screen_size(void *lcd)
{
  return static_cast<Lcd_jz4740_chip *>(lcd)->get_screen_size();
}

l4_addr_t
jz4740_lcd_get_palette(void *lcd, l4_addr_t base)
{
  return static_cast<Lcd_jz4740_chip *>(lcd)->get_palette(base);
}

void
jz4740_lcd_init_palette(void *lcd, l4_addr_t palette)
{
  static_cast<Lcd_jz4740_chip *>(lcd)->init_palette(palette);
}
